from dataclasses import dataclass
import numpy as np

@dataclass
class QuadPoint():
    """Quadtree point.

    A dataclass for storing a point value at coordinates (x, y).
    
    Args:
        _x (float): x position.
        _y (float): y position.
        _value (float, optional): Point value. Defaults to -inf.
    """
    _x: float
    _y: float
    _value: float = -np.inf
    
    @property
    def x(self) -> float:
        return self._x
    @x.setter
    def x(self, val: float) -> None:
        self._x = val
        
    @property
    def y(self) -> float:
        return self._y
    @y.setter
    def y(self, val: float) -> None:
        self._y = val
        
    @property
    def value(self) -> float:
        return self._value
    @value.setter
    def value(self, val: float) -> None:
        self._value = val


import pytest
import numpy as np

from astroqtpy.quadnode import QuadNode
from astroqtpy.quadpoint import QuadPoint

def test_nbody_node() -> None:
    """Test the QuadNode class
    
    """
    # test instantiation of Nbody Node
    my_node = QuadNode(0, 1, 0, 1)
    with pytest.raises(ValueError):
        my_node = QuadNode(0, -1, 0, 1)
    with pytest.raises(ValueError):
        my_node = QuadNode(0, 1, 0, -1)
    with pytest.raises(ValueError):
        my_node = QuadNode(0, 0, 0, 1)
    with pytest.raises(ValueError):
        my_node = QuadNode(0, 1, 0, 0)
    with pytest.raises(ValueError):
        my_node = QuadNode(-np.inf, 1, 0, 1)
    with pytest.raises(ValueError):
        my_node = QuadNode(0, 1, 0, np.inf)
    
    # test node splitting
    assert my_node._is_split() == False
    my_node.split_node()
    assert my_node._is_split() == True
    
    # test node value generation
    my_node.node_points.append(QuadPoint(0.5, 0.5, 12))
    my_node.node_points.append(QuadPoint(0.25, 0.75, 15))
    my_node.node_points.append(QuadPoint(0.1, 0.6, 16))
    my_node._generate_node_value('mean')  # should be 14.333334
    assert my_node.get_node_value('mean') == pytest.approx(14.333334)
    with pytest.raises(ValueError):
        my_node._generate_node_value('fake_statistic')  # should throw value error
    
    
if __name__ == "__main__":
    test_nbody_node()


from astroqtpy.quadpoint import QuadPoint

def test_quad_point() -> None:
    """Test the QuadPoint dataclass.
    
    """

    # set ground truth values
    x = 2.3
    y = 3.1
    value = 8.3
    
    # new instanace of QuadPoint with ground truth values
    point = QuadPoint(x, y, value)
    assert point.x == x
    assert point.y == y
    assert point.value == value
    
    # new ground truth values
    x_new = 100.3
    y_new = -4.6
    value_new = 0.0
    
    # set new ground truth values
    point.x = x_new
    point.y = y_new
    point.val = value_new
    assert point.x == x_new
    assert point.y == y_new
    assert point.val == value_new
    
    
    
if __name__ == "__main__":
    test_quad_point()


import sys
import abc

import numpy as np
from matplotlib import axes, cm, colors
from rebound.interruptible_pool import InterruptiblePool  # import throws `pkg_resources.declare_namespace` warning

from .quadnode import QuadNode
from .quadpoint import QuadPoint


class BaseTree(abc.ABC):
    """Base quadtree.

    The abstract base class for all types of astroQTpy quadtrees.

    Args:
        x_min (float): Minimum x value for this quadtree.
        x_max (float): Maximum x value for this quadtree.
        y_min (float): Minimum y value for this quadtree.
        y_max (float): Maximum y value for this quadtree.
        split_threshold (float, optional): Threshold discrepancy in order to split nodes. Defaults to 0.2.
        node_statistic (str, optional): Statistic to compute node values ['mean', 'std', or 'median']. Defaults to 'mean'.
        N_points (int, optional): Maximum number of points per node. Defaults to 20.
        min_depth (int, optional): Minimum quadtree depth. Defaults to 3.
        max_depth (int, optional):  Maximum quadtree depth. Defaults to 6.
        N_proc (int, optional): Number of cores for multiprocessing. Defaults to 1.
        verbose (bool, optional): Option to print node values in real time. Defaults to False.
        filename_points (str, optional): Name of output file to save points. Defaults to 'points.txt'.
        filename_nodes (str, optional): Name of output file to save nodes. Defaults to 'nodes.txt'.
    """

    def __init__(self,
        x_min: float,
        x_max: float,
        y_min: float,
        y_max: float,
        split_threshold: float = 0.2,
        node_statistic: str = 'mean',
        N_points: int = 20,
        min_depth: int = 3,
        max_depth: int = 6,
        N_proc: int = 1,
        verbose: bool = False,
        filename_points: str = 'points.txt',
        filename_nodes: str = 'nodes.txt'
        ) -> None:
        """__init__

        Create the base quadtree class for astroQTpy.
        """
        
        super().__init__()
        
        # check user input attributes
        if split_threshold <= 0:
            raise ValueError('split_threshold must be greater than zero.')
        elif not np.isfinite(split_threshold):
            raise ValueError('split_threshold must be finite.')
        else:
            self.split_threshold = split_threshold
            
        if node_statistic in ['mean', 'median', 'std']:
            self.node_statistic = node_statistic
        else:
            raise ValueError('node_statistic must be either "mean, "median", or "std"')
        
        if N_points <= 0:
            raise ValueError('N_points must be greater than zero.')
        else:
            self.N_points = N_points
            
        if max_depth < min_depth:
            raise ValueError('max_depth must be greater than min_depth.')
        else:
            self.min_depth = min_depth
            self.max_depth = max_depth
            
        if N_proc <= 0:
            raise ValueError('N_proc must be greater than zero.')
        else:
            self.N_proc = N_proc
        
        # define other attributes
        self.verbose = verbose
        self.filename_points = filename_points
        self.filename_nodes = filename_nodes
        self.node_count = 1
        self.root = QuadNode(x_min, x_max, y_min, y_max, 1)
        self.min_node_value = np.inf  # for plotting limits
        self.max_node_value = -np.inf
        
    
    def _compare_nodes(self, northwest: QuadNode, southeast: QuadNode, dir_northsouth: bool = False) -> None:
        """Compare nodes.

        Compare values of two nodes and split into 4 children if threshold is exceeded.

        Args:
            northwest (QuadNode): First quadtree node.
            southeast (QuadNode): Second quadtree node.
            dir_northsouth (bool, optional): Whether the nodes being compared are aligned N-S. Defaults to False.
        """
        
        # if both northwest and southeast nodes are split
        if northwest._is_split() and southeast._is_split():
            if dir_northsouth:
                self._compare_nodes(northwest.child_sw, southeast.child_nw, True)
                self._compare_nodes(northwest.child_se, southeast.child_ne, True)
            else:
                self._compare_nodes(northwest.child_ne, southeast.child_nw, False)
                self._compare_nodes(northwest.child_se, southeast.child_sw, False)
                
        # if northwest is split and southeast is NOT split
        elif northwest._is_split() and not southeast._is_split():
            if dir_northsouth:
                self._compare_nodes(northwest.child_sw, southeast, True)
                self._compare_nodes(northwest.child_se, southeast, True)
            else:
                self._compare_nodes(northwest.child_ne, southeast, False)
                self._compare_nodes(northwest.child_se, southeast, False)
            
        # if northwest is NOT split and southeast is split
        elif not northwest._is_split() and southeast._is_split():
            if dir_northsouth:
                self._compare_nodes(northwest, southeast.child_nw, True)
                self._compare_nodes(northwest, southeast.child_ne, True)
            else:
                self._compare_nodes(northwest, southeast.child_nw, False)
                self._compare_nodes(northwest, southeast.child_sw, False)
                
        # if neither node is split
        else:
            if abs(northwest.get_node_value(self.node_statistic) - southeast.get_node_value(self.node_statistic)) >= self.split_threshold:
                if northwest.depth >= southeast.depth and southeast.depth < self.max_depth:
                    southeast.split_node()
                    self.fill(southeast, self.N_points)
                    self._save_checkpoint()
                    
                if southeast.depth >= northwest.depth and northwest.depth < self.max_depth:
                    northwest.split_node()
                    self.fill(northwest, self.N_points)
                    self._save_checkpoint()
                    
                    
    def fill(self, node: QuadNode, N_points: int) -> None:
        """Fill.

        Fill a given node with points.

        Args:
            node (QuadNode): The quadtree node to be filled.
            N_points (int): Number of points to put inside this node.
        """
        if node._is_split():
            self.fill(node.child_nw, N_points)
            self.fill(node.child_ne, N_points)
            self.fill(node.child_sw, N_points)
            self.fill(node.child_se, N_points)
            return
        
        if self.N_proc > 1:
            self.evaluate_multiple_points(node, N_points)

        else:
            while len(node.node_points) < N_points:
                point = self.evaluate_point(node, rng_seed=np.random.randint(1, 1e8))
                node.node_points.append(point)
     
     
    def evaluate_multiple_points(self, node: QuadNode, N_points: int) -> None:
        """Evaulate multiple points.

        Evaluate multiple points contained within a given node in parallel.

        Args:
            node (QuadNode): Node in which to evaulate points.
            N_points (int): Maximum number of points to evaulate within this node.
        """
        N_empty = int(N_points - len(node.node_points))
        map_iters = [(node, np.random.randint(1, 1e8)) for _ in range(N_empty)]
        
        with InterruptiblePool(processes=self.N_proc) as pool:
            points = pool.starmap(self.evaluate_point, map_iters)
            for point in points:
                node.node_points.append(point)
                
                
    @abc.abstractmethod
    def evaluate_point(self, node: QuadNode, rng_seed: int = 123456) -> QuadPoint:
        """Evaluate point.

        Abstract method to calculate the value of one point within a given node.

        Args:
            node (QuadNode): Node in which to evaulate point.
            rng_seed (int, optional): Random number generator seed. Defaults to 123456.

        Returns:
            QuadPoint: A QuadPoint object that has been evaluated.
        """
        pass
    
    
    def _save_checkpoint(self) -> None:
        """Save checkpoint.
        
        Convenience function to save progress and update node count (called any time a node is split).
        """
        self.node_count = self.node_count + 3
        self.print_all_points()
        self.print_all_nodes()
        if self.verbose:
            print(f"Progress saved. (nodes = {self.node_count})")
        
    
    def print_all_points(self) -> None:
        """Print all points.
        
        Print all current quadtree points to a file ('filename_points').
        """
        stdout_ = sys.stdout
        f = open(self.filename_points, 'w')
        sys.stdout = f
        
        print("# x \t y \t value")
        self.root.print_node_points()
        sys.stdout = stdout_
        
        f.close()
        
    
    def print_all_nodes(self) -> None:
        """Print all nodes.
        
        Print all current quadtree node values to a file ('filename_nodes').
        """
        stdout_ = sys.stdout
        f = open(self.filename_nodes, 'w')
        sys.stdout = f
        
        print("# Depth \t x_min \t x_max \t y_min \t y_max \t node value")
        self.root.print_node_value(self.node_statistic)
        sys.stdout = stdout_
        
        f.close()
    
    
    def load_points(self) -> None:
        """Load points.
        
        Load all points from a previously saved quadtree.
        """
        with open(self.filename_points, 'r') as f:
            for line in f:
                if line[0] == "#":
                    continue
                line_spl = line.split('\t')
                point = QuadPoint(float(line_spl[0]), float(line_spl[1]), float(line_spl[2]))
                self.root.node_points.append(point)
        
        self.squeeze_node(self.root)
        
    
    def squeeze_node(self, node: QuadNode) -> None:
        """Squeeze node.
        
        Distribute quadtree points from parent to children.
        """
        if len(node.node_points) > self.N_points and node.depth < self.max_depth:
            node.split_node()
            self.node_count = self.node_count + 3
        
        if node._is_split():
            self.squeeze_node(node.child_nw)
            self.squeeze_node(node.child_ne)
            self.squeeze_node(node.child_sw)
            self.squeeze_node(node.child_se)
            
            
    def run_quadtree(self) -> None:
        """Run quadtree.
        
        Run the quadtree from a previously saved run, or start a new run.
        """
        # attempt to load previous results
        try:
            print("Attempting to load previous results...")
            self.load_points()
            print(f"   {self.node_count} nodes found, starting from previous checkpoint...")
        except FileNotFoundError:
            print("   No previous results found, starting new...")

        # must execute at least minimum depth nodes
        for _ in range(self.min_depth):
            self._forward(self.root)

        # save
        self._save_checkpoint()
        
        print("DONE! :)")
        
        
    def _forward(self, node: QuadNode) -> None:
        """Forward.
        
        Advance quadtree forward a step by comparing child nodes and expanding resolution where necessary.
        
        Args:
            node (QuadNode): Quadtree node.
        """
        if node._is_split():
            self._compare_nodes(node.child_nw, node.child_ne, False)
            self._compare_nodes(node.child_sw, node.child_se, False)
            self._compare_nodes(node.child_nw, node.child_sw, True)
            self._compare_nodes(node.child_ne, node.child_se, True)
        elif node.depth < self.min_depth:
            node.split_node()
            self._save_checkpoint()
            
        if node._is_split():
            self._forward(node.child_nw)
            self._forward(node.child_ne)
            self._forward(node.child_sw)
            self._forward(node.child_se)
        elif len(node.node_points) < self.N_points:
            if self.N_proc > 1:
                self.evaluate_multiple_points(node, self.N_points)
            else:
                while len(node.node_points) < self.N_points:
                    point = self.evaluate_point(node, rng_seed=np.random.randint(1, 1e8))
                    node.node_points.append(point)
                    
                    
    def _get_min_max_nodes(self, node: QuadNode) -> None:
        """Get min max nodes.

        Convenience function to grab the minimum and maximum node values for plotting.

        Args:
            node (QuadNode): Starting quadtree node.
        """
        if node._is_split():
            self._get_min_max_nodes(node.child_nw)
            self._get_min_max_nodes(node.child_ne)
            self._get_min_max_nodes(node.child_sw)
            self._get_min_max_nodes(node.child_se)
        
        if node.get_node_value(self.node_statistic) > self.max_node_value:
            self.max_node_value = node.get_node_value(self.node_statistic)
            
        if node.get_node_value(self.node_statistic) < self.min_node_value and node.get_node_value(self.node_statistic) != -np.inf:
            self.min_node_value = node.get_node_value(self.node_statistic)
                                
    
    def _draw_nodes(self, ax: axes.Axes, node: QuadNode, mappable: cm.ScalarMappable,
                    show_lines: bool, show_points: bool, show_values: bool) -> None:
        """Draw nodes.
        
        Convenience function to plot quadtree nodes, including children.
        
        Args:
            ax (:obj:`matplotlib.axes.Axes`): Axis for plotting.
            node (QuadNode): Quadtree node.
            mappable (:obj:`matplotlib.cm.ScalarMappable`): Scalar mappable for color mapping.
            show_lines (bool): Whether to draw boundary lines between nodes.
            show_points (bool): Whether to plot points contained within this node.
            show_values (bool): Whether to print the value of this node on the plot.
        """
        
        if node._is_split():
            self._draw_nodes(ax, node.child_nw, mappable, show_lines, show_points, show_values)
            self._draw_nodes(ax, node.child_ne, mappable, show_lines, show_points, show_values)
            self._draw_nodes(ax, node.child_sw, mappable, show_lines, show_points, show_values)
            self._draw_nodes(ax, node.child_se, mappable, show_lines, show_points, show_values)
            
        node.draw_node(ax, mappable, show_lines, show_points, show_values, self.node_statistic)
        
    
    def draw_tree(self, ax: axes.Axes,
                  cmap: str = 'RdYlGn_r',
                  vmin: float = None,
                  vmax: float = None,
                  show_lines: float = True,
                  show_points: bool = False,
                  show_values: bool = False
                  ) -> cm.ScalarMappable:
        """Draw quadtree. 
        
        Plot the entire quadtree on a given axis.

        Args:
            ax (:obj:`matplotlib.axes.Axes`): Matplotlib axis for plotting.
            cmap (str, optional): Matplotlib colormap. Defaults to 'RdYlGn_r'.
            vmin (float, optional): Minimum value for colorbar. Defaults to None.
            vmax (float, optional): Maximum value for colorbar. Defaults to None.
            show_lines (float, optional): Option to plot node boundary lines. Defaults to True.
            show_points (bool, optional): Option to plot node points. Defaults to False.
            show_values (bool, optional): Option to print node values on plot. Defaults to False.

        Returns:
            matplotlib.cm.ScalarMappable: Matplotlib ScalarMappable.
        """
        if vmin is None:
            self._get_min_max_nodes(self.root)
            vmin = 0.8 * self.min_node_value
        if vmax is None:
            self._get_min_max_nodes(self.root)
            vmax = 1.2 * self.max_node_value
        
        mappable = cm.ScalarMappable(
            colors.Normalize(vmin, vmax),
            cmap=cmap
            )
        
        self._draw_nodes(ax, self.root, mappable, show_lines, show_points, show_values)
        
        return mappable



import numpy as np 
from matplotlib import axes, cm

class QuadNode():
    """Quadtree node.

    A class for astroQTpy quadtree nodes. Each node is defined by its given boundary limits
    and 'depth', and may contain up to any given number of quadtree points.
    
    Args:
        x_min (float): Minimum x value of this node.
        x_max (float): Maximum x value of this node.
        y_min (float): Minimum y value of this node.
        y_max (float): Maximum y value of this node.
        depth (int, optional): Depth of this node. Defaults to 0.
    """
    def __init__(self,
        x_min : float,
        x_max : float,
        y_min : float,
        y_max : float,
        depth : int = 0
        ) -> None:
        """__init__

        Create a quadtree node for astroQTpy.
        """        
        # check user specified attributes
        if x_max < x_min:
            raise ValueError('x_max must be greater than x_min.')
        elif x_max == x_min:
            raise ValueError('x_max cannot equal x_min.')
        elif not np.isfinite(x_min) or not np.isfinite(x_max):
            raise ValueError('x limits must be finite.')
        else:
            self.x_min = x_min
            self.x_max = x_max
        
        if y_max < y_min:
            raise ValueError('y_max must be greater than y_min.')
        elif y_max == y_min:
            raise ValueError('y_max cannot equal y_min.')
        elif not np.isfinite(y_max) or not np.isfinite(y_min):
            raise ValueError('y limits must be finite.')
        else:
            self.y_min = y_min
            self.y_max = y_max
        
        if depth < 0:
            raise ValueError('depth cannot be negative.')
        elif not np.isfinite(depth):
            raise ValueError('depth must be finite.')
        else:
            self.depth = depth
        
        # define other attributes
        self.node_value = -np.inf
        self.node_points = []  # to store QuadPoint objects
        self.child_nw = None
        self.child_ne = None
        self.child_sw = None
        self.child_se = None
     
    
    def split_node(self) -> None:
        """Split node.

        Split this node into 4 equal 'child' nodes. Distribute any points
        contained within this node to its children.
        """
        x_center = 0.5 * (self.x_min + self.x_max)
        y_center = 0.5 * (self.y_min + self.y_max)
        
        # create children
        self.child_nw = QuadNode(self.x_min, x_center, y_center, self.y_max, self.depth + 1)
        self.child_ne = QuadNode(x_center, self.x_max, y_center, self.y_max, self.depth + 1)
        self.child_sw = QuadNode(self.x_min, x_center, self.y_min, y_center, self.depth + 1)
        self.child_se = QuadNode(x_center, self.x_max, self.y_min, y_center, self.depth + 1)
        
        # distribute completed trials from parent node to child nodes
        for parent_point in self.node_points:
            # NW child
            if self.child_nw.x_min < parent_point.x < self.child_nw.x_max and \
               self.child_nw.y_min < parent_point.y < self.child_nw.y_max:
                self.child_nw.node_points.append(parent_point)
            
            # NE child
            if self.child_ne.x_min < parent_point.x < self.child_ne.x_max and \
               self.child_ne.y_min < parent_point.y < self.child_ne.y_max:
                self.child_ne.node_points.append(parent_point)
                
            # SW child
            if self.child_sw.x_min < parent_point.x < self.child_sw.x_max and \
               self.child_sw.y_min < parent_point.y < self.child_sw.y_max:
                self.child_sw.node_points.append(parent_point)
                
            # SE child
            if self.child_se.x_min < parent_point.x < self.child_se.x_max and \
               self.child_se.y_min < parent_point.y < self.child_se.y_max:
                self.child_se.node_points.append(parent_point)
        
        # clear parent node points
        self.node_points.clear()
    
    
    def _generate_node_value(self, statistic: str) -> None:
        """Generate node value.
        
        Calculate an aggragate value of all points contained within this node.
        
        Args:
            statistic (str): Statistic to compute for this node. Choose from ['mean', 'std', or 'median'].
        """
                
        if len(self.node_points) == 0:
            return
        
        node_point_values = [point.value for point in self.node_points]
        
        if statistic == 'mean':
            self.node_value = np.mean(node_point_values)
        elif statistic == 'std':
            self.node_value = np.std(node_point_values)
        elif statistic == 'median':
            self.node_value = np.median(node_point_values)
        else:
            raise ValueError(" Node statistic must be either 'mean', 'std', or 'median'. ")
    
    
    def get_node_value(self, statistic: str) -> float:
        """Get node value.
        
        Convenience function to grab this node's value.
        
        Args:
            statistic (str): Statistic to pass to '_generate_node_value'. Choose from ['mean', 'std', or 'median'].

        Returns:
            float: Node value.
        """
        if self.node_value == -np.inf:
            self._generate_node_value(statistic)
            
        return self.node_value
    
    
    def _is_split(self) -> None:
        """Is split.
        
        Convenience function to check whether this node has been split.
        """
        return self.child_nw is not None
    
    
    def print_node_points(self) -> None:
        """Print node points.
        
        Convenience function to print each point contained within this node to a given file.
        """
        if self._is_split():
            self.child_nw.print_node_points()
            self.child_ne.print_node_points()
            self.child_sw.print_node_points()
            self.child_se.print_node_points()
            
        else:
            print(f"# Depth = {self.depth}, x = {self.x_min:.5f} - {self.x_max:.5f}, " 
                  f"y = {self.y_min:.5f} - {self.y_max:.5f}")
            
            for point in self.node_points:
                print(f"{point.x:.5f}\t{point.y:.5f}\t{point.value}\t")
          
          
    def print_node_value(self, statistic: str) -> None:
        """Print node value.
        
        Convenience function to print the value of this node to a given file.
        
        Args:
            statistic (str): Statistic to pass to 'get_node_value'. Choose from ['mean', 'std', or 'median'].
        """
        if self._is_split():
            self.child_nw.print_node_value(statistic)
            self.child_ne.print_node_value(statistic)
            self.child_sw.print_node_value(statistic)
            self.child_se.print_node_value(statistic)
            
        else:
            print(f"{self.depth}\t{self.x_min:.5f}\t{self.x_max:.5f}\t" 
                  f"{self.y_min:.5f}\t{self.y_max:.5f}\t{self.get_node_value(statistic):.3f}\t")
    
    
    def draw_node(self, ax: axes.Axes, mappable: cm.ScalarMappable,
                  show_lines: bool, show_points: bool, show_values: bool,
                  statistic: str) -> None:
        """Draw node.

        Plot this node on a matplotlib axis.

        Args:
            ax (:obj:`matplotlib.axes.Axes`): Axis for plotting.
            mappable (:obj:`matplotlib.cm.ScalarMappable`): Scalar mappable for color mapping.
            show_lines (bool): Whether to draw boundary lines between nodes.
            show_points (bool): Whether to plot points contained within this node.
            show_values (bool): Whether to print the value of this node on the plot.
            statistic (str): Statistic to pass to 'get_node_value'. Choose from ['mean', 'std', or 'median'].
        """
        # grab node limits for convenience        
        x1, x2 = self.x_min, self.x_max
        y1, y2 = self.y_min, self.y_max
        
        if not self._is_split():
            ax.fill_between([x1,x2], [y1,y1], [y2,y2], color=mappable.to_rgba(self.get_node_value(statistic)))
            
            if show_lines:
                ax.plot([x1, x2, x2, x1, x1], [y2, y2, y1, y1, y2],
                        c="k", lw=1, alpha=0.5
                        )
            
            if show_points:
                if len(self.node_points) > 0:
                    ax.scatter(*zip(*[(point.x, point.y) for point in self.node_points]),
                            c='k', s=1, marker='.', alpha=0.8, rasterized=True
                            )
                
            if show_values:
                x_mid = 0.5 * (x1 + x2)
                y_mid = 0.5 * (y1 + y2)
                ax.text(x_mid, y_mid, round(self.get_node_value(statistic), 2),
                        horizontalalignment="center", verticalalignment="center", c="k", size=10
                        )


from setuptools import setup, find_packages
import re

# auto-updating version code stolen from orbitize! stolen from RadVel
def get_property(prop, project):
    result = re.search(
        r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop),
        open(project + "/__init__.py").read(),
    )
    return result.group(1)

def get_requires():
    reqs = []
    for line in open("requirements.txt", "r").readlines():
        reqs.append(line)
    return reqs

setup(
    name="astroQTpy",
    version=get_property("__version__", "astroqtpy"),
    description="astroQTpy implements a quadtree data structure to explore 2D parameter space",
    url="https://github.com/CalebHarada/astroQTpy",
    author="Caleb K. Harada",
    author_email="charad@berkeley.edu",
    license="MIT",
    packages=find_packages(),
    keywords="Quadtree Astronomy",
    install_requires=get_requires(),
)


import numpy as np

from astroqtpy.quadnode import QuadNode

from .quadnode import QuadNode
from .quadpoint import QuadPoint
from .basetree import BaseTree


class RandomQuadTree(BaseTree):
    """Random quadtree.

    A class for creating a quadtree with randomly sampled points.

    Args:
        x_min (float): Minimum x value for this quadtree.
        x_max (float): Maximum x value for this quadtree.
        y_min (float): Minimum y value for this quadtree.
        y_max (float): Maximum y value for this quadtree.
        split_threshold (float, optional): Threshold discrepancy in order to split nodes. Defaults to 0.2.
        node_statistic (str, optional): Statistic to compute node values ['mean', 'std', or 'median']. Defaults to 'mean'.
        N_points (int, optional): Maximum number of points per node. Defaults to 20.
        min_depth (int, optional): Minimum quadtree depth. Defaults to 3.
        max_depth (int, optional):  Maximum quadtree depth. Defaults to 6.
        N_proc (int, optional): Number of cores for multiprocessing. Defaults to 1.
        verbose (bool, optional): Option to print node values in real time. Defaults to False.
        filename_points (str, optional): Name of output file to save points. Defaults to 'points.txt'.
        filename_nodes (str, optional): Name of output file to save nodes. Defaults to 'nodes.txt'.
    """
    
    def __init__(self,
                 x_min: float,
                 x_max: float,
                 y_min: float,
                 y_max: float,
                 split_threshold: float = 0.2,
                 node_statistic: str = 'mean',
                 N_points: int = 20,
                 min_depth: int = 3,
                 max_depth: int = 6,
                 N_proc: int = 1,
                 verbose: bool = False,
                 filename_points: str = 'points.txt',
                 filename_nodes: str = 'nodes.txt'
                 ) -> None:
        """__init__

        Create this random quadtree.
        """
        super().__init__(x_min,
                         x_max,
                         y_min,
                         y_max,
                         split_threshold,
                         node_statistic,
                         N_points,
                         min_depth,
                         max_depth,
                         N_proc,
                         verbose,
                         filename_points,
                         filename_nodes
                         )
        
        
    def evaluate_point(self, node: QuadNode, rng_seed: int = 123456) -> QuadPoint:
        """Evaluate point.

        Calculate the value of one point within a given node as either 1 or 0.

        Args:
            node (QuadNode): Quadtree node in which to evaulate point.
            rng_seed (int, optional): Random number generator seed. Defaults to 123456.

        Returns:
            QuadPoint: A QuadPoint object.
        """
        rng = np.random.default_rng(rng_seed)
        
        point = QuadPoint(
            rng.uniform(node.x_min, node.x_max),
            rng.uniform(node.y_min, node.y_max),
            rng.choice([1, 0])
            )
    
        return point





class Chi2QuadTree(BaseTree):
    """Chi2 quadtree.
    
    A quadtree class for mapping 2-parameter reduced :math:`\\chi^2` values.
    
    .. math::
        \\chi^2_r = \\frac{\\chi^2}{\\nu}
        
    where :math:`\\nu` is the number of degrees of freedom and
    
    .. math::
        \\chi^2 = \\sum_{i=0}^{N-1} \\frac{\\Big(y_i - f(x_i | a, b)\\Big)^2}{\\sigma_i^2}
        
    .. note::
    
        The `model_func` argument must be callable and take two arguments: `x_data` and parameters `(a, b)`.
        It must return a single float value `y`. For example:
        
        .. code-block:: python
        
            def my_model(x, parameters):
                a, b = parameters  # unpack parameters
                
                # ... do something ...
                y = a * x + b
                
                return y
                
        See tutorials for more examples.
        
    Args:
        x_min (float): Minimum x value (param `a`) for this quadtree.
        x_max (float): Maximum x value (param `a`)for this quadtree.
        y_min (float): Minimum y value (param `b`)for this quadtree.
        y_max (float): Maximum y value (param `b`)for this quadtree.
        data (:obj:`np.ndarray`): Data array. Must have shape (2, N).
        model_func (callable): Function :math:`f(x | a, b)` to calculate a model to compare to data.
        weights (:obj:`np.ndarray`, optional): Data weights, typically expressed as :math:`1/\\sigma^2`. Defaults to None.
        max_chi2 (float, optional): Largest permitted reduced :math:`\\chi^2`. Defaults to 10.
        split_threshold (float, optional): Threshold discrepancy in order to split nodes. Defaults to 0.2.
        node_statistic (str, optional): Statistic to compute node values ['mean', 'std', or 'median']. Defaults to 'mean'.
        N_points (int, optional): Maximum number of points per node. Defaults to 20.
        min_depth (int, optional): Minimum quadtree depth. Defaults to 3.
        max_depth (int, optional):  Maximum quadtree depth. Defaults to 6.
        N_proc (int, optional): Number of cores for multiprocessing. Defaults to 1.
        verbose (bool, optional): Option to print node values in real time. Defaults to False.
        filename_points (str, optional): Name of output file to save points. Defaults to 'points.txt'.
        filename_nodes (str, optional): Name of output file to save nodes. Defaults to 'nodes.txt'.
    """
    
    def __init__(self,
                 x_min: float,
                 x_max: float,
                 y_min: float,
                 y_max: float,
                 data: np.ndarray,
                 model_func: callable,
                 weights: np.ndarray = None,
                 max_chi2: float = 10,
                 split_threshold: float = 0.2,
                 node_statistic: str = 'mean',
                 N_points: int = 20,
                 min_depth: int = 3,
                 max_depth: int = 6,
                 N_proc: int = 1,
                 verbose: bool = False,
                 filename_points: str = 'points.txt',
                 filename_nodes: str = 'nodes.txt'
                 ) -> None:
        """__init__

        Create a chi-squared quadtree.
        """
        super().__init__(x_min,
                         x_max,
                         y_min,
                         y_max,
                         split_threshold,
                         node_statistic,
                         N_points,
                         min_depth,
                         max_depth,
                         N_proc,
                         verbose,
                         filename_points, 
                         filename_nodes)
        
        # check inputs
        if not type(data) == np.ndarray:
            raise TypeError('data must be ndarray.')
        elif np.shape(data)[0] != 2:
            raise ValueError('data array must have shape (2, N).')
        else:
            self.data = data
        
        if not callable(model_func):
            raise TypeError('model_func must be callable.')
        else:
            self.model_func = model_func
            
        if weights is not None:
            if len(weights) != np.shape(data)[1]:
                raise ValueError('length of weights array must be equal to data.')
            else:
                self.weights = weights
        else:
            self.weights = np.ones(np.shape(data)[1])
            
        if max_chi2 <= 0:
            raise ValueError('max_chi2 must be greater than zero.')
        else:
            self.max_chi2 = max_chi2
            
        # define other attributes
        self.chi2_min_point = QuadPoint(0, 0, np.inf)  # just a dummy point for now
        
    
    def evaluate_point(self, node: QuadNode, rng_seed: int = 123456) -> QuadPoint:
        """Evaluate point.

        Calculate the value of one point within a given node.

        Args:
            node (QuadNode): Quadtree node in which to evaulate point.
            rng_seed (int, optional): Random number generator seed. Defaults to 123456.

        Returns:
            QuadPoint: A QuadPoint object.
        """
        rng = np.random.default_rng(rng_seed)
        _x = rng.uniform(node.x_min, node.x_max)
        _y = rng.uniform(node.y_min, node.y_max)
        
        # grab data
        x_data = self.data[0]
        y_data = self.data[1]
        
        # calculate model
        model_params = (_x, _y)
        y_model = self.model_func(x_data, model_params)
        
        # calculate chi-squared
        dof = len(y_data) - 2
        chi2 = np.sum(self.weights * (y_data - y_model)**2)
        
        # reduced chi-square
        chi2_r = np.min((self.max_chi2, chi2 / dof))  # if really large, set to max_chi2
        
        point = QuadPoint(_x, _y, chi2_r)
    
        return point
    
    
    def _find_chi2_min(self, node: QuadNode) -> None:
        """Find chi^2 min.

        Convenience function to locate the point with the smallest chi^2 value.

        Args:
            node (QuadNode): Quadtree node.
        """
        
        if node._is_split():
            self._find_chi2_min(node.child_nw)
            self._find_chi2_min(node.child_ne)
            self._find_chi2_min(node.child_sw)
            self._find_chi2_min(node.child_se)
        
        for point in node.node_points:
            if point.value < self.chi2_min_point.value:
                self.chi2_min_point = point
            
    
    def get_chi2_min(self) -> QuadPoint:
        """Get chi^2 min.

        Get the point with the smallest reduced chi^2 value.
        
        Returns:
            QuadPoint: Quadtree point with least chi^2.
        """
        
        if not np.isfinite(self.chi2_min_point.value):
            self._find_chi2_min(self.root)
            
        return self.chi2_min_point
        
        
        





class NbodyQuadTree(BaseTree):
    """Nbody quadtree.

    A class for creating a quadtree for running N-body simulations.
    
    .. note::
    
        The `simulation_func` argument must be callable and take a single argument `(x, y)`.
        It must return a single float value. For example:
        
        .. code-block:: python
        
            def my_function(parameters):
                x, y = parameters  # unpack parameters
                
                # ... do something ...
                number = float(1)
                
                return number
                
        See tutorials for more examples.


    Args:
        x_min (float): Minimum x value for this quadtree.
        x_max (float): Maximum x value for this quadtree.
        y_min (float): Minimum y value for this quadtree.
        y_max (float): Maximum y value for this quadtree.
        simulation_func (callable): Function to calculate the outcome of an Nbody simulation.
        split_threshold (float, optional): Threshold discrepancy in order to split nodes. Defaults to 0.2.
        node_statistic (str, optional): Statistic to compute node values ['mean', 'std', or 'median']. Defaults to 'mean'.
        N_points (int, optional): Maximum number of points per node. Defaults to 20.
        min_depth (int, optional): Minimum quadtree depth. Defaults to 3.
        max_depth (int, optional):  Maximum quadtree depth. Defaults to 6.
        N_proc (int, optional): Number of cores for multiprocessing. Defaults to 1.
        verbose (bool, optional): Option to print node values in real time. Defaults to False.
        filename_points (str, optional): Name of output file to save points. Defaults to 'points.txt'.
        filename_nodes (str, optional): Name of output file to save nodes. Defaults to 'nodes.txt'.
    """
    def __init__(self,
                 x_min: float,
                 x_max: float,
                 y_min: float,
                 y_max: float,
                 simulation_func: callable,
                 split_threshold: float = 0.2,
                 node_statistic: str = 'mean',
                 N_points: int = 20,
                 min_depth: int = 3,
                 max_depth: int = 6,
                 N_proc: int = 1,
                 verbose: bool = False,
                 filename_points: str = 'points.txt',
                 filename_nodes: str = 'nodes.txt'
                 ) -> None:
        """__init__

        Create an N-body simulation quadtree.
        """
        super().__init__(x_min,
                         x_max,
                         y_min,
                         y_max,
                         split_threshold,
                         node_statistic,
                         N_points,
                         min_depth,
                         max_depth,
                         N_proc,
                         verbose,
                         filename_points,
                         filename_nodes)
        
        # check that input function is callable
        if not callable(simulation_func):
            raise TypeError('simulation_func must be callable.')
        else:
            self.simulation_func = simulation_func
        
            
    def evaluate_point(self, node: QuadNode, rng_seed: int = 123456) -> QuadPoint:
        """Evaluate point.

        Calculate the value of one point within a given node from an N-body simulation.

        Args:
            node (QuadNode): Quadtree node in which to evaulate point.
            rng_seed (int, optional): Random number generator seed. Defaults to 123456.

        Returns:
            QuadPoint: A QuadPoint object.
        """
        rng = np.random.default_rng(rng_seed)
        _x = rng.uniform(node.x_min, node.x_max)
        _y = rng.uniform(node.y_min, node.y_max)
        
        # run N-body sim
        sim = self.simulation_func((_x, _y))
        
        point = QuadPoint(_x, _y, sim)
    
        return point

        
        
        
        
        